# =============================================================================
# Casanova Reader
# =============================================================================
#
# A fast but comfortable CSV reader based upon csv.reader to avoid dealing
# with csv.DictReader which is nice but very slow.
#
import csv

from casanova.utils import is_contiguous, ensure_open
from casanova.exceptions import EmptyFileError, MissingColumnError


class HeadersPositions(object):
    def __init__(self, headers):
        if isinstance(headers, int):
            self.__headers = list(range(headers))
            self._dict = {i: i for i in self.__headers}
        else:
            self.__headers = headers
            self._dict = {h: i for i, h in enumerate(self.__headers)}

            for name, value in self._dict.items():
                if name.isidentifier():
                    setattr(self, name, value)

    def __len__(self):
        return len(self.__headers)

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict


def get_column_index(pos, key, default=None):
    try:
        return pos[key]
    except KeyError:
        return default


def collect_column_indices(pos, columns):
    indices = []

    for column in columns:
        i = get_column_index(pos, column)

        if i is None:
            raise MissingColumnError

        indices.append(i)

    return indices


class CasanovaReader(object):
    namespace = 'casanova.reader'

    def __init__(self, input_file, no_headers=False, encoding='utf-8'):

        # Should we open a file for the user?
        input_file = ensure_open(input_file, encoding=encoding)

        self.input_file = input_file
        self.reader = csv.reader(input_file)
        self.fieldnames = None
        self.first_row = None
        self.can_slice = True
        self.binary = False

        if no_headers:
            try:
                self.first_row = next(self.reader)
            except StopIteration:
                raise EmptyFileError

            self.pos = HeadersPositions(len(self.first_row))
        else:
            try:
                self.fieldnames = next(self.reader)
            except StopIteration:
                raise EmptyFileError

            self.pos = HeadersPositions(self.fieldnames)

    def __repr__(self):
        columns_info = ' '.join('%s=%s' % t for t in zip(self.pos._fields, self.pos))

        return '<%s %s>' % (self.namespace, columns_info)

    def iter(self):
        if self.first_row is not None:
            yield self.first_row
            self.first_row = None

        yield from self.reader

    def __iter__(self):
        return self.iter()

    def __records(self, columns, with_rows=False):
        pos = collect_column_indices(self.pos, columns)

        if self.can_slice and is_contiguous(pos):
            if len(pos) == 1:
                s = slice(pos[0], pos[0] + 1)
            else:
                s = slice(pos[0], pos[1] + 1)

            if with_rows:
                def iterator():
                    for row in self.iter():
                        yield row, row[s]
            else:
                def iterator():
                    for row in self.iter():
                        yield row[s]
        else:
            if with_rows:
                def iterator():
                    for row in self.iter():
                        yield row, [row[i] for i in pos]
            else:
                def iterator():
                    for row in self.iter():
                        yield [row[i] for i in pos]

        return iterator()

    def __cells(self, column, with_rows=False):
        i = get_column_index(self.pos, column)

        if i is None:
            raise MissingColumnError(column)

        if with_rows:
            def iterator():
                for row in self.iter():
                    yield row, row[i]
        else:
            def iterator():
                for row in self.iter():
                    yield row[i]

        return iterator()

    def cells(self, column, with_rows=False):
        if not isinstance(column, (str, int)):
            return self.__records(column, with_rows=with_rows)

        return self.__cells(column, with_rows=with_rows)

    def close(self):
        self.input_file.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @classmethod
    def count(cls, input_file, max_rows=None, **kwargs):
        assert max_rows is None or max_rows > 0, '%s.count: expected max_rows to be `None` or > 0.' % cls.namespace

        n = 0

        with cls(input_file, **kwargs) as reader:
            for _ in reader:
                n += 1

                if max_rows is not None and n > max_rows:
                    return None

        return n



# =============================================================================
# Casanova Monkey Reader
# =============================================================================
#
# A Casanova reader relying on csvmonkey for performance.
#
import csvmonkey

from casanova.reader import CasanovaReader, HeadersPositions
from casanova.utils import is_binary_buffer, ensure_open
from casanova.exceptions import InvalidFileError, EmptyFileError


class CasanovaMonkeyReader(CasanovaReader):
    namespace = 'casanova_monkey.reader'

    def __init__(self, input_file, no_headers=False, encoding='utf-8', lazy=False):

        # Should we open a file for the user?
        input_file = ensure_open(input_file, encoding=encoding, mode='rb')

        # Ensuring we are reading a binary buffer
        if not is_binary_buffer(input_file):
            raise InvalidFileError('casanova_monkey.reader: expecting file in binary mode (e.g. "rb") or BytesIO.')

        self.input_file = input_file
        self.reader = csvmonkey.from_file(input_file, header=False)
        self.fieldnames = None
        self.first_row = None
        self.can_slice = not lazy
        self.binary = True
        self.lazy = lazy

        if no_headers:
            try:
                self.first_row = next(self.reader)
            except StopIteration:
                raise EmptyFileError

            self.pos = HeadersPositions(len(self.first_row))
        else:
            try:
                self.fieldnames = next(self.reader).aslist()
            except StopIteration:
                raise EmptyFileError

            self.pos = HeadersPositions(self.fieldnames)

    def iter(self):
        if self.first_row is not None:
            if self.lazy:
                yield self.first_row
            else:
                yield self.first_row.aslist()
            self.first_row = None

        if self.lazy:
            yield from self.reader
        else:
            for row in self.reader:
                yield row.aslist()



# =============================================================================
# Casanova Monkey Library Endpoint
# =============================================================================
#
from casanova_monkey.enricher import (
    CasanovaMonkeyEnricher as enricher,
    ThreadsafeCasanovaMonkeyEnricher as threadsafe_enricher
)
from casanova_monkey.reader import (
    CasanovaMonkeyReader as reader
)



# =============================================================================
# Casanova Reverse Reader
# =============================================================================
#
# A reader reading the file backwards in order to read last lines in constant
# time. This is sometimes useful to be able to resume some computations
# where they were left off.
#
import csv
from io import DEFAULT_BUFFER_SIZE
from file_read_backwards.file_read_backwards import FileReadBackwardsIterator

from casanova.reader import CasanovaReader
from casanova.utils import lookahead, ensure_open
from casanova.exceptions import EmptyFileError


class CasanovaReverseReader(CasanovaReader):
    namespace = 'casanova.reverse_reader'

    def __init__(self, input_file, **kwargs):
        super().__init__(input_file, **kwargs)

        self.backwards_file = ensure_open(self.input_file.name, mode='rb')

        backwards_iterator = FileReadBackwardsIterator(
            self.backwards_file,
            self.input_file.encoding,
            DEFAULT_BUFFER_SIZE
        )

        backwards_reader = csv.reader(backwards_iterator)

        def generator():

            for row, has_more in lookahead(backwards_reader):
                if has_more or self.fieldnames is None:
                    yield row

            self.close()

        self.reader = generator()

        if self.fieldnames is None:
            self.first_row = None

    def close(self):
        super().close()
        self.backwards_file.close()

    @staticmethod
    def last_cell(input_file, column, **kwargs):
        with CasanovaReverseReader(input_file, **kwargs) as reader:
            try:
                for record in reader.cells(column):
                    return record
            except StopIteration:
                raise EmptyFileError



# =============================================================================
# Casanova Exceptions
# =============================================================================
#


class CasanovaError(Exception):
    pass


class ColumnNumberMismatchError(CasanovaError):
    pass


class EmptyFileError(CasanovaError):
    pass


class MissingColumnError(CasanovaError):
    pass


class InvalidFileError(CasanovaError):
    pass


class NotResumableError(CasanovaError):
    pass


class ResumeError(CasanovaError):
    pass


class CorruptedIndexColumn(CasanovaError):
    pass



# =============================================================================
# Casanova Library Endpoint
# =============================================================================
#
from casanova.contiguous_range_set import ContiguousRangeSet
from casanova.enricher import (
    CasanovaEnricher as enricher,
    ThreadsafeCasanovaEnricher as threadsafe_enricher
)
from casanova.reader import (
    CasanovaReader as reader
)
from casanova.reverse_reader import (
    CasanovaReverseReader as reverse_reader
)



# =============================================================================
# Casanova Monkey Enricher
# =============================================================================
#
# A Casanova enricher relying on csvmonkey for performance.
#
from casanova_monkey.reader import CasanovaMonkeyReader
from casanova.enricher import make_enricher

ThreadsafeCasanovaMonkeyEnricher, CasanovaMonkeyEnricher = make_enricher(
    'CasanovaMonkeyEnricher',
    'casanova_monkey.enricher',
    CasanovaMonkeyReader
)



# =============================================================================
# Casanova Utils
# =============================================================================
#
# Miscellaneous utility functions.
#
import gzip
import sys
import codecs
from io import BytesIO, BufferedReader, TextIOWrapper


def iter_with_prev(iterator):
    prev_item = None

    for item in iterator:
        if prev_item is not None:
            yield prev_item, item

        prev_item = item


def lookahead(iterator):

    try:
        last = next(iterator)
    except StopIteration:
        return

    for item in iterator:
        yield last, True
        last = item

    yield last, False


def is_contiguous(l):
    for p, n in iter_with_prev(l):
        if p != n - 1:
            return False

    return True


def is_binary_buffer(buf):
    if isinstance(buf, gzip.GzipFile) and not isinstance(buf, TextIOWrapper):
        return True
    if isinstance(buf, BufferedReader):
        if 'b' not in buf.mode:
            return False
    elif not isinstance(buf, BytesIO):
        return False

    return True


def is_resumable_buffer(buf):
    if not isinstance(buf, (BufferedReader, TextIOWrapper)):
        return False

    if 'a' not in buf.mode and '+' not in buf.mode:
        return False

    return True


def is_empty_buffer(buf):
    return buf.tell() == 0


def is_mute_buffer(buf):
    return buf is sys.stdout or buf is sys.stderr or not hasattr(buf, 'tell')


def encoding_fingerprint(encoding):
    return encoding.lower().replace('-', '')


def ensure_open(p, encoding='utf-8', mode='r'):
    if not isinstance(p, str):
        return p

    if p.endswith('.gz'):
        if 'b' in mode:
            return gzip.open(p, mode=mode)

        mode += 't'
        return gzip.open(p, encoding=encoding, mode=mode)

    if encoding_fingerprint(encoding) != 'utf8':
        return codecs.open(p, encoding=encoding, mode=mode)

    return open(p, mode=mode)



# =============================================================================
# Casanova Enricher
# =============================================================================
#
# A CSV reader/writer combo that can be used to read an input CSV file and
# easily ouput a similar CSV file while editing, adding and filtering cell_count.
#
import os
import csv
from threading import Lock

from casanova.contiguous_range_set import ContiguousRangeSet
from casanova.exceptions import (
    NotResumableError,
    ResumeError,
    MissingColumnError,
    CorruptedIndexColumn
)
from casanova.reader import (
    CasanovaReader,
    get_column_index,
    collect_column_indices
)
from casanova.utils import (
    is_resumable_buffer,
    is_empty_buffer,
    is_mute_buffer
)


def make_enricher(name, namespace, Reader):

    class AbstractCasanovaEnricher(Reader):
        __name__ = name

        def __init__(self, input_file, output_file, no_headers=False,
                     resumable=False, auto_resume=True, keep=None, add=None,
                     listener=None, prepend=None):

            # Inheritance
            super().__init__(
                input_file,
                no_headers=no_headers
            )

            # Sanity tests
            if resumable and not is_resumable_buffer(output_file):
                raise NotResumableError('%s: expecting an "a+" or "a+b" buffer.' % namespace)

            self.output_file = output_file
            self.writer = csv.writer(output_file)
            self.keep_indices = None
            self.output_fieldnames = self.fieldnames
            self.added_count = 0
            self.padding = None

            self.resumable = resumable
            self.should_resume = False
            self.already_done_count = 0

            self.listener = listener

            if keep is not None:
                self.keep_indices = collect_column_indices(self.pos, keep)
                self.output_fieldnames = self.filterrow(self.output_fieldnames)

            if add is not None:
                self.output_fieldnames += add
                self.added_count = len(add)
                self.padding = [''] * self.added_count

            if prepend is not None:
                self.output_fieldnames = prepend + self.output_fieldnames

            # Need to write headers?
            output_buffer_is_empty = is_mute_buffer(output_file) or is_empty_buffer(output_file)

            if not no_headers:

                if not resumable or output_buffer_is_empty:
                    self.writeheader()

            # Resuming
            if resumable and not output_buffer_is_empty:
                self.should_resume = True

                if auto_resume:
                    self.resume()

        def __repr__(self):
            columns_info = ' '.join('%s=%s' % t for t in zip(self.pos._fields, self.pos))

            return '<%s%s%s %s>' % (
                namespace,
                ' resumable' if self.resumable else '',
                ' unordered' if getattr(self, 'unordered', False) else '',
                columns_info
            )

        def resume(self):

            if not self.should_resume:
                return

            self.should_resume = False

            # Rolling back to beginning of file
            output_file = self.output_file

            if self.binary:
                output_file = open(output_file.name, 'rb')
            else:
                output_file.seek(0, os.SEEK_SET)

            reader = Reader(output_file, no_headers=self.fieldnames is None)

            should_emit = callable(self.listener)

            if should_emit:
                self.listener('resume.start', None)

            for row in reader:
                self.already_done_count += 1

                if should_emit:
                    self.listener('resume.output', row)

            if self.binary:
                output_file.close()

            i = 0

            while i < self.already_done_count:
                try:
                    row = next(self.reader)

                    if should_emit:
                        self.listener('resume.input', row)

                    i += 1
                except StopIteration:
                    raise ResumeError('%s.resume: output has more lines than input.' % namespace)

        def filterrow(self, row):
            if self.keep_indices is not None:
                row = [row[i] for i in self.keep_indices]

            return row

        def formatrow(self, row, add=None, index=None):

            # Additions
            if self.added_count > 0:
                if add is None:
                    add = self.padding
                else:
                    assert len(add) == self.added_count, '%s.writerow: expected %i additional cells but got %i.' % (namespace, self.added_count, len(add))

                row = self.filterrow(row) + add

            # No additions
            else:
                assert add is None, '%s.writerow: expected no additions.' % namespace

                row = self.filterrow(row)

            if index is not None:
                row = [index] + row

            return row

        def writeheader(self):
            self.writer.writerow(self.output_fieldnames)

        def writerow(self, row, add=None):
            self.writer.writerow(self.formatrow(row, add))

    class AbstractThreadsafeCasanovaEnricher(AbstractCasanovaEnricher):
        __name__ = 'Threadsafe' + name

        def __init__(self, input_file, output_file, no_headers=False,
                     resumable=False, auto_resume=True, keep=None, add=None,
                     listener=None, index_column='index'):

            self.index_column = index_column
            self.event_lock = Lock()
            self.already_done = ContiguousRangeSet()

            # Inheritance
            super().__init__(
                input_file,
                output_file,
                no_headers=no_headers,
                resumable=resumable,
                keep=keep,
                add=add,
                listener=listener,
                prepend=[index_column]
            )

        def __iter__(self):
            iterator = enumerate(super().__iter__())
            should_emit = callable(self.listener)

            for index, row in iterator:
                if self.already_done.stateful_contains(index):
                    if should_emit:
                        with self.event_lock:
                            self.listener('resume.input', row)

                    continue

                yield index, row

        def resume(self):

            # Rolling back to beginning of file
            output_file = self.output_file

            if self.binary:
                output_file = open(output_file.name, 'rb')
            else:
                output_file.seek(0, os.SEEK_SET)

            reader = Reader(output_file, no_headers=self.fieldnames is None)

            should_emit = callable(self.listener)

            i = get_column_index(reader.pos, self.index_column)

            if i is None:
                raise MissingColumnError(self.index_column)

            for row in reader:
                try:
                    current_index = int(row[i])
                except ValueError:
                    raise CorruptedIndexColumn

                self.already_done.add(current_index)

                if should_emit:
                    with self.event_lock:
                        self.listener('resume.output', row)

            if self.binary:
                output_file.close()

        def cells(self, column, with_rows=False):
            if with_rows:
                index = 0
                for row, value in super().cells(column, with_rows=True):
                    yield index, row, value
                    index += 1
            else:
                yield from enumerate(super().cells(column))

        def writerow(self, index, row, add=None):
            self.writer.writerow(self.formatrow(row, add, index=index))

    return AbstractThreadsafeCasanovaEnricher, AbstractCasanovaEnricher


ThreadsafeCasanovaEnricher, CasanovaEnricher = make_enricher(
    'CasanovaEnricher',
    'casanova.enricher',
    CasanovaReader
)



from setuptools import setup, find_packages

with open('./README.md', 'r') as f:
    long_description = f.read()

setup(name='casanova',
      version='0.8.0',
      description='Specialized & performant CSV readers, writers and enrichers for python.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/medialab/casanova',
      license='MIT',
      author='Guillaume Plique',
      author_email='kropotkinepiotr@gmail.com',
      keywords='csv',
      python_requires='>=3.5',
      packages=find_packages(exclude=['benchmark', 'test']),
      package_data={'docs': ['README.md']},
      install_requires=[
          'file-read-backwards>=2.0.0'
      ],
      extras_require={
          'monkey': ['csvmonkey==0.0.5']
      },
      zip_safe=True)
